/*
     This file is part of GNUnet.
     (C) 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/testbed/commands.c 
 * @brief the commands available in the testbed
 * @author Ronaldo Alves Ferreira
 * @author Christian Grothoff
 * @author Murali Krishan Ramanathan
 *
 * Todo:
 * - implement more messages (primarily AFS control)
 * - test
 **/

#include "testbed.h"
#include "commands.h"
#include "socket.h"
#include "get-stats.h"

/**
 * @brief struct keeping per-peer information for the testbed
 */
typedef struct {
  /** IP address of the peer */
  IPaddr	     ip;
  /** CS port of the peer */
  unsigned short     port;
  /** string describing the peer address */
  char	* ips;
  /** socket to communicate with the peer */
  GNUNET_TCP_SOCKET sock;
  /** HELO message identifying the peer in the network */
  HELO_Message * helo;
} NODE_INFO;

/**
 * List of nodes known to the testbed.
 */
static NODE_INFO * nodes = NULL;

/**
 * Number of known nodes, size of the nodes array.
 */
static int nnodes = 0;

/**
 * Should the driver exit?
 */
int do_quit = NO;

/**
 * Convert the strings ss and ds to peer-identifiers (ints) s and d
 * respectively.  Aborts the method containing the macro with an error
 * message if the peer-IDs are not within range.
 */
#define CHECK_SRC_DST(s, d, ss, ds)	\
  s = atoi(ss);			\
  d = atoi(ds);				    \
  if (s < 0 || s >= nnodes || d < 0 || d >= nnodes) {  \
    PRINTF("Invalid src (%s) or dst (%s)\n", ss, ds);  \
    return -1;					       \
  }

/**
 * Convert the string ps to a peer-id p and abort the current method
 * with an error message if ps is not a valid peer-id.
 */
#define CHECK_PEER(p, ps)	\
  p = atoi(ps);				\
  if (p < 0 || p >= nnodes) {			\
    PRINTF("Invalid peer value %s\n", ps);	\
    return -1;					\
  }

/**
 * Send a message to peer 'peer' of type 'msgType'
 * the given size and data.
 */
static int sendMessage(unsigned msgType, 
		       int peer, 
		       unsigned short argSize, 
		       void *arg) {
  TESTBED_CS_MESSAGE * msg;
  int msgsz;
  
  /* Assume peer value is valid. */
  if (argSize + sizeof(TESTBED_CS_MESSAGE) < 65535) 
    errexit("Message body too big for sendMessage: %s\n",
	    argSize);
  
  msgsz = sizeof(TESTBED_CS_MESSAGE)+argSize;
  msg = MALLOC(msgsz);
  msg->header.size 
    = htons(msgsz);
  msg->header.tcpType 
    = htons(TESTBED_CS_PROTO_REQUEST);
  msg->msgType
    = htonl(msgType);
  memcpy(&msg->data[0], 
	 arg, 
	 argSize);
  msgsz = writeToSocket(&nodes[peer].sock,
			&msg->header);
  FREE(msg);
  if (msgsz == SYSERR) {
    PRINTF("ERROR: Could not send message to peer %s.\n", 
	   nodes[peer].ips);
    return SYSERR;
  }
  return OK;
}

/**
 * Read a result from the given peer.  Print
 * an error message if the peer fails to respond.
 *
 * @return OK on success, SYSERR on error
 */
static int readResult(int peer,
		      int * result) {
  if (OK != readTCPResult(&nodes[peer].sock,
			  result)) {
    PRINTF("ERROR: peer %s is not responding.\n",
	   nodes[peer].ips);
    return SYSERR;
  }
  return OK;
}

/* ****************** individual commands ********** */


/**
 * Add a node to the configuration.
 * Arguments must be IP PORT of the peer.
 */
static int addNode(int argc, char * argv[]) {
  int currindex;
  TESTBED_HELO_MESSAGE * hdr;
  
  if (argc != 2) {
    PRINTF("Syntax: add-node IP PORT.\n");
    return -1;
  }
  /* connect */
  currindex = nnodes;
  GROW(nodes, nnodes, nnodes+1);
  nodes[currindex].ips = STRDUP(argv[0]);
  nodes[currindex].port = atoi(argv[1]);
#ifndef MINGW
  inet_aton(argv[0], &nodes[currindex].ip.addr);
#else
  nodes[currindex].ip.addr.S_un.S_addr = inet_addr(argv[0]);
#endif
  
  if (SYSERR == initGNUnetClientSocket(nodes[currindex].port,
				       nodes[currindex].ips,
				       &nodes[currindex].sock)) {
    PRINTF("ERROR: could not connect to %s:%d.\n",
	   nodes[currindex].ips,
	   nodes[currindex].port);
    return -1;
  }
  
  /* request HELO */
  if (OK != sendMessage(TESTBED_GET_HELO,
			currindex,
			0,
			NULL)) {
    PRINTF("ERROR: Could not send message to %s.\n",
	   nodes[currindex].ips);
    FREE(nodes[currindex].ips);
    GROW(nodes, 
	 nnodes, 
	 nnodes-1);
    return -1;
  }
  
  hdr = NULL;
  if (SYSERR == readFromSocket(&nodes[currindex].sock, 
			       (CS_HEADER**)&hdr)) {
    PRINTF("ERROR: peer %s is not responding.\n",
	   nodes[currindex].ips);
    FREE(nodes[currindex].ips);
    GROW(nodes, nnodes, nnodes-1);
    return -1;
  }
  nodes[currindex].helo 
    = MALLOC(HELO_Message_size(&hdr->helo));
  memcpy(nodes[currindex].helo, 
	 &hdr->helo, 
	 HELO_Message_size(&hdr->helo));
  FREE(hdr);
  PRINTF("%d\n",
	 currindex);
  return 0;
}

/**
 * Tear down the connection between two peers.
 */
static int delConnection(int argc,
			 char * argv[]) {
  int src, dst, ack;
  
  if (argc != 2) {
    PRINTF("Syntax: disconnect PEERID PEERID\n");
    return -1;
  }
  CHECK_SRC_DST(src, dst, argv[0], argv[1]);
  if (OK != sendMessage(TESTBED_DEL_PEER, 
			src,
			sizeof(HostIdentity),
			&nodes[dst].helo->senderIdentity)) 
    return -1;
  if (OK != readResult(src,
		       &ack))
    return -1;
  if (ack == OK) {
    PRINTF("OK.\n");
    return 0;
  } else {
    PRINTF("ERROR: Connection NOT deleted.\n");
    return -1;
  }
}

/**
 * Tear down all connections of a peer.
 */
static int delAllConnections(int argc,
			     char * argv[]) {
  int dst, ack;
  
  if (argc != 1) {
    PRINTF("Syntax: disconnect-all PEERID\n");
    return -1;
  }
  CHECK_PEER(dst, argv[0]);
  if (OK != sendMessage(TESTBED_DEL_ALL_PEERS, 
			dst,
			0,
			NULL)) 
    return -1;
  if (OK != readResult(dst,
		       &ack))
    return -1;
  if (ack == OK) {
    PRINTF("OK.\n");
    return 0;
  } else {
    PRINTF("ERROR: Connections NOT deleted.\n");
    return -1;
  }
}

/**
 * Add a connection between two peers.
 */
static int addConnection(int argc, 
			 char * argv[]) {
  int src, dst, ack;
  
  if (argc != 2) {
    PRINTF("Syntax: connect PEERID PEERID\n");
    return -1;
  }
  CHECK_SRC_DST(src, dst, argv[0], argv[1]);
  if (SYSERR == sendMessage(TESTBED_ADD_PEER,
			    src,
			    HELO_Message_size(nodes[dst].helo),
			    nodes[dst].helo)) 
    return -1;
  if (OK != readResult(src,
		       &ack))
    return -1;  
  if (ack == OK) {
    PRINTF("OK.\n");
    return 0;
  } else {
    PRINTF("ERROR: peer can not connect.\n");
    return -1;
  }
}

/**
 * Set the level of trust that one peer has in
 * another.
 */
static int setTrust(int argc, char * argv[]) {
  int src, dst, value, ack;
  TESTBED_SET_TVALUE_MESSAGE msg;
  
  if (argc != 3) {
    PRINTF("Syntax: set-trust PEERID PEERID TRUST\n");
    return -1;
  }
  CHECK_SRC_DST(src, dst, argv[0], argv[1]);
  value = atoi(argv[2]);
  msg.trust = htonl(value);
  memcpy(&msg.otherPeer,
	 &nodes[dst].helo->senderIdentity,
	 sizeof(HostIdentity));
  if (SYSERR == sendMessage(TESTBED_SET_TVALUE,
			    src,
			    sizeof(HostIdentity)+sizeof(unsigned int),
			    &msg.otherPeer)) 
    return -1;  
  if (OK != readResult(src,
		       &ack)) 
    return -1;
  if (htonl(ack) != OK) {
    PRINTF("ERROR: peer could not set trust value.\n");
    return -1;
  } else {
    PRINTF("OK.\n");
    return 0;
  }
}

/**
 * Get the amount of trust that A has in B.
 */
static int getTrust(int argc, char *argv[]) {
  int src, dst, value;
  
  if (argc != 2) {
    PRINTF("Syntax: get-trust PEERID PEERID\n");
    return -1;
  }
  CHECK_SRC_DST(src, dst, argv[0], argv[1]);
  if (SYSERR == sendMessage(TESTBED_GET_TVALUE, 
			    src, 
			    sizeof(HostIdentity),
			    &nodes[dst].helo->senderIdentity)) 
    return -1;
  if (SYSERR == readResult(src, &value)) 
    return -1;
  if (value < 0) {
    PRINTF("ERROR: could not get trust value.\n");
    return -1;
  } else {
    PRINTF("%d\n", 
	   value);
    return 0;
  }
}

/**
 * Disable HELO at a peer.
 */
static int disableHELO(int argc, char *argv[]) {
  int dst, value;
  
  if (argc != 2) {
    PRINTF("Syntax: helo-disable PEERID\n");
    return -1;
  }
  CHECK_PEER(dst, argv[0]);
  if (SYSERR == sendMessage(TESTBED_DISABLE_HELO, 
			    dst, 
			    0,
			    NULL)) 
    return -1;
  if (SYSERR == readResult(dst, &value)) 
    return -1;
  if (value != OK) {
    PRINTF("ERROR: could disable HELO\n");
    return -1;
  } else {
    PRINTF("OK.\n");
    return 0;
  }
}

/**
 * Enable HELO at a peer.
 */
static int enableHELO(int argc, char *argv[]) {
  int dst, value;
  
  if (argc != 2) {
    PRINTF("Syntax: helo-enable PEERID\n");
    return -1;
  }
  CHECK_PEER(dst, argv[0]);
  if (SYSERR == sendMessage(TESTBED_ENABLE_HELO, 
			    dst, 
			    0,
			    NULL)) 
    return -1;
  if (SYSERR == readResult(dst, &value)) 
    return -1;
  if (value != OK) {
    PRINTF("ERROR: could enable HELO\n");
    return -1;
  } else {
    PRINTF("OK.\n");
    return 0;
  }
}

/**
 * Disable AUTOCONNECT at a peer.
 */
static int disableAUTOCONNECT(int argc, char *argv[]) {
  int dst, value;
  
  if (argc != 2) {
    PRINTF("Syntax: helo-disable PEERID\n");
    return -1;
  }
  CHECK_PEER(dst, argv[0]);
  if (SYSERR == sendMessage(TESTBED_DISABLE_AUTOCONNECT, 
			    dst, 
			    0,
			    NULL)) 
    return -1;
  if (SYSERR == readResult(dst, &value)) 
    return -1;
  if (value != OK) {
    PRINTF("ERROR: could disable AUTOCONNECT\n");
    return -1;
  } else {
    PRINTF("OK.\n");
    return 0;
  }
}

/**
 * Enable AUTOCONNECT at a peer.
 */
static int enableAUTOCONNECT(int argc, char *argv[]) {
  int dst, value;
  
  if (argc != 2) {
    PRINTF("Syntax: helo-enable PEERID\n");
    return -1;
  }
  CHECK_PEER(dst, argv[0]);
  if (SYSERR == sendMessage(TESTBED_ENABLE_AUTOCONNECT, 
			    dst, 
			    0,
			    NULL)) 
    return -1;
  if (SYSERR == readResult(dst, &value)) 
    return -1;
  if (value != OK) {
    PRINTF("ERROR: could enable AUTOCONNECT\n");
    return -1;
  } else {
    PRINTF("OK.\n");
    return 0;
  }
}


static int allowDenyConnectHelper(unsigned int argc,
				  char * argv[],
				  int type) {
  int dst, value;
  HostIdentity * list;
  int i;
  int idx = 0;
  
  CHECK_PEER(dst, argv[0]);
  if (argc > (65532 - sizeof(TESTBED_CS_MESSAGE)) / sizeof(HostIdentity)) {
    PRINTF("Too many peers specified.  Ask a wizzard to enlarge limit.\n");
    return -1;
  }

  list = NULL;
  for (i=1;i<argc;i++) 
    CHECK_PEER(idx, argv[i]); /* may return, do before MALLOC! */
  if (argc > 1)
    list = MALLOC(sizeof(HostIdentity)*(argc-1));
  for (i=1;i<argc;i++) {
    CHECK_PEER(idx, argv[i]);
    memcpy(&list[i-1],
	   &nodes[idx].helo->senderIdentity,
	   sizeof(HostIdentity));
  }
  if (SYSERR == sendMessage(type,
			    dst, 
			    sizeof(HostIdentity)*(argc-1),
			    list)) {
    FREENONNULL(list);
    return -1;  
  }
  FREENONNULL(list);
  if (SYSERR == readResult(dst, &value)) 
    return -1;
  if (value != OK) {
    PRINTF("ERROR: could change setting.\n");
    return -1;
  } else {
    PRINTF("OK.\n");
    return 0;
  }
}

/**
 * Deny connections to certain peers at a peer.
 */
static int denyConnect(int argc, char *argv[]) {
  if (argc < 1) {
    PRINTF("Syntax: connect-deny PEERID [PEERID]*\n");
    return -1;
  }
  return allowDenyConnectHelper(argc, argv, 
				TESTBED_DENY_CONNECT);
}

/**
 * Allow connections to certain peers at a peer.
 */
static int allowConnect(int argc, char *argv[]) {
  if (argc < 1) {
    PRINTF("Syntax: connect-allow PEERID [PEERID]*\n");
    return -1;
  }
  return allowDenyConnectHelper(argc, argv,
				TESTBED_ALLOW_CONNECT);
}

/**
 * Helper function for (un)loadModule.
 * @param type load or unload requested?
 */
static int loadModuleHelper(unsigned short type, 
			    char * peerId,
			    char * modulename) {
  int ok, dst;
  
  CHECK_PEER(dst, peerId);
  if (OK != sendMessage(type, 
			dst, 
			strlen(modulename), 
			modulename)) 
    return -1;  
  if (OK != readResult(dst, 
		       &ok)) 
    return -1;
  if (ok != OK) {
    PRINTF("ERROR: peer %s refused.\n",
	   nodes[dst].ips);
    return -1;
  }
  PRINTF("OK.\n");
  return 0;
}

/**
 * Load an application module at the given peer.
 */
static int loadModule(int argc, char *argv[]) {
  if (argc != 2) {
    PRINTF("Syntax: load-module PEERID MODULENAME\n");
    return -1;
  }
  return loadModuleHelper(TESTBED_LOAD_MODULE, 
			  argv[0],
			  argv[1]);
}

/**
 * Unload an application module.
 */
static int unloadModule(int argc, char *argv[]) {
  if (argc != 2) {
    PRINTF("Syntax: unload-module PEERID MODULENAME\n");
    return -1;
  }
  return loadModuleHelper(TESTBED_UNLOAD_MODULE,
			  argv[0],
			  argv[1]);
}

/**
 * Fork a client process.  Captures the output and makes it
 * available via process-output.  The client can be killed
 * using process-signal.  The process identifier is printed.
 */
static int startProcess(int argc, char *argv[]) {
  char * cmdLine;
  int size;
  int i;
  int dst;
  int ack;

  if (argc < 2) {
    PRINTF("Syntax: process-start PEERID COMMAND [ARGUMENTS]\n");
    return -1;
  }
  CHECK_PEER(dst, argv[0]);
  size = 0;
  for (i=1;i<argc;i++)
    size += 1 + strlen(argv[i]);
  cmdLine = MALLOC(size);
  cmdLine[0] = '\0';
  for (i=1;i<argc;i++)
    strcat(cmdLine, argv[i]);
  
  if (OK != sendMessage(TESTBED_EXEC, 
			dst,
			size,
			cmdLine))
    return -1;
  if (OK != readResult(dst,
		       &ack))
    return -1;
  if (ack != SYSERR) {
    PRINTF("%d\n",
	   ack);
    return 0;
  } else {
    PRINTF("ERROR: Peer could not fork process.\n");
    return -1;
  }  
}

/**
 * Send a signal to a client process.  Use signal
 * 0 to stop recording process output.
 */
static int signalProcess(int argc, char *argv[]) {
  int dst;
  int ack;
  TESTBED_SIGNAL_MESSAGE msg;

  if (argc != 3) {
    PRINTF("Syntax: process-signal PEERID PROCESSID SIGNAL\n");
    return -1;
  }
  CHECK_PEER(dst, argv[0]);
  msg.pid = htonl(atoi(argv[1]));
  msg.signal = htonl(atoi(argv[2]));
  if (OK != sendMessage(TESTBED_SIGNAL, 
			dst,
			sizeof(TESTBED_SIGNAL_MESSAGE) - sizeof(TESTBED_CS_MESSAGE),
			&msg.pid))
    return -1;
  if (OK != readResult(dst,
		       &ack))
    return -1;
  if (ack == OK) {
    PRINTF("OK.\n");
    return 0;
  } else {
    PRINTF("ERROR: Peer could not signal process.\n");
    return -1;
  }  
}

/**
 * Get the recorded output of a process.
 */
static int dumpProcessOutput(int argc, 
			     char * argv[]) {
  int dst;
  int pid;
  int ack;

  if (argc != 2) {
    PRINTF("Syntax: process-output PEERID PROCESSID\n");
    return -1;
  }
  CHECK_PEER(dst, argv[0]);
  pid = htonl(atoi(argv[1]));
  if (OK != sendMessage(TESTBED_GET_OUTPUT, 
			dst,
			sizeof(int),
			&pid))
    return -1;
  if (OK != readResult(dst,
		       &ack))
    return -1;
  if (ack != SYSERR) {
    int pos = 0;
    while (pos < ack) {
      unsigned short size;
      TESTBED_OUTPUT_REPLY_MESSAGE * reply;   

      reply = NULL;
      if (SYSERR == readFromSocket(&nodes[dst].sock, 
				   (CS_HEADER**)&reply)) {
	PRINTF("ERROR: peer %s is not responding after %d of %d bytes.\n",
	       nodes[dst].ips,
	       pos,
	       ack);
	return -1;
      }
      /* FIXME: check that this is the correct reply format */
      size = ntohs(reply->header.header.size);
      PRINTF("%*.s", 
	     size,
	     reply->data);
      FREE(reply);
      pos += size;      
    }    
    return 0;
  } else {
    PRINTF("ERROR: Peer could not return process output.\n");
    return -1;
  }  
}

/**
 * Set bandwidth limitations for a peer.
 */
static int setBW(int argc, char * argv[]) {
  TESTBED_SET_BW_MESSAGE msg;
  int dst, in, out, ack;
  
  if (argc != 3) {
    PRINTF("Syntax: set-bw PEERID DOWN-BPS UP-BPS\n");
    return -1;
  }
  CHECK_PEER(dst, argv[0]);
  in  = atoi(argv[1]);
  out = atoi(argv[2]);
  if ((in < 0) || (out < 0)) {
    PRINTF("ERROR: Invalid bandwidth specification.\n");
    return -1;
  }
  msg.in_bw  = htonl(in);
  msg.out_bw = htonl(out);  
  if (SYSERR == sendMessage(TESTBED_SET_BW,
			    dst,
			    sizeof(TESTBED_SET_BW_MESSAGE) - sizeof(TESTBED_CS_MESSAGE),
			    &msg.in_bw)) 
    return -1;
  if (OK != readResult(dst, &ack)) 
    return -1;
  if (ack != OK) {
    PRINTF("ERROR: peer could not set the specified bandwith.\n");
    return -1;
  } else {
    PRINTF("OK.\n");
    return 0;
  }
}

/**
 * Set artifical message loss rates for a peer.
 */
static int setLoss(int argc, char * argv[]) {
  int dst;
  int ack;
  TESTBED_SET_LOSS_RATE_MESSAGE msg;

  if (argc != 3) {
    PRINTF("Syntax: set-loss PEERID DOWN-LOSS UP-LOSS\n");
    return -1;
  }  
  CHECK_PEER(dst, argv[0]);
  msg.percentageLossInbound 
    = htonl(atoi(argv[1]));
  msg.percentageLossOutbound 
    = htonl(atoi(argv[2]));
  
  if (SYSERR == sendMessage(TESTBED_SET_LOSS_RATE,
			    dst,
			    sizeof(TESTBED_SET_LOSS_RATE_MESSAGE) - sizeof(TESTBED_CS_MESSAGE),
			    &msg.percentageLossInbound))
    return -1;
  if (OK != readResult(dst, &ack)) 
    return -1;
  if (ack != OK) {
    PRINTF("ERROR: peer could not set the specified loss rates.\n");
    return -1;
  } else {
    PRINTF("OK.\n");
    return 0;
  }
}

/**
 * Obtain statistics from a peer.
 *
 * @param argc number of arguments from the command line
 * @param argv command line arguments
 * @return 0 ok, 1 on error
 **/   
static int getStat(int argc, char ** argv) {
  int res, peer, printProtocols;
  
  printProtocols = NO;
  if (argc != 2) {
    PRINTF("Syntax: get-stat PEERID STATID\n");
    return -1;
  }
  CHECK_PEER(peer, argv[0]);
  res = requestAndPrintStatistic(&nodes[peer].sock,
				 argv[1]);
  if (res == OK)
    return 0;
  else
    return -1;
}

/**
 * Obtain statistics from a peer.
 *
 * @param argc number of arguments from the command line
 * @param argv command line arguments
 * @return 0 ok, 1 on error
 **/   
static int getStats(int argc, char ** argv) {
  int res, peer, printProtocols;
  
  printProtocols = NO;
  if (argc == 2) {
    if (strcmp(argv[0], "-P")) {
      PRINTF("Syntax: get-stats [-P] PEERID\n");
      return -1;
    }
    printProtocols = YES;
    CHECK_PEER(peer, argv[1]);
  } else if (argc != 1) {
    PRINTF("Syntax: get-stats [-P] PEERID\n");
    return -1;
  } else
    CHECK_PEER(peer, argv[0]);
  res = requestAndPrintStatistics(&nodes[peer].sock);
  if ( (printProtocols == YES) && 
       (res == OK)) {
    res = requestAndPrintProtocols(&nodes[peer].sock);
  }
  if (res == OK)
    return 0;
  else
    return -1;
}


/**
 * Upload a file to a peer.
 */
static int uploadFile(int argc, 
		      char *argv[]) {
  int peer, nbytes, flen, ack;
  char * buf;
  FILE * infile;
  TESTBED_UPLOAD_FILE_MESSAGE * msg;
    
  if (argc != 3) {
    PRINTF("Syntax: load-file PEERID LOCAL_FILENAME DEST_FILENAME\n");
    return -1;
  }
  CHECK_PEER(peer, argv[0]);
  infile = FOPEN(argv[1], "r");
  if (infile == NULL) {
    PRINTF("ERROR: Could not open file %s\n",
	   argv[1]);
    return -1;
  }
  flen = strlen(argv[2]) + 1; /* '\0' added in the flen */
  if (flen > TESTBED_FILE_BLK_SIZE) {
    PRINTF("ERROR: destination file name too long (%d characters, limit %d).\n", 
	   flen-1,
	   TESTBED_FILE_BLK_SIZE);
    return -1;
  }

  msg = MALLOC(sizeof(TESTBED_UPLOAD_FILE_MESSAGE) + TESTBED_FILE_BLK_SIZE);
  msg->header.header.size 
    = htons(sizeof(TESTBED_UPLOAD_FILE_MESSAGE)+flen);
  msg->header.header.tcpType 
    = htons(TESTBED_CS_PROTO_REQUEST);
  msg->header.msgType
    = htonl(TESTBED_UPLOAD_FILE);
  msg->type 
    = htonl(TESTBED_FILE_DELETE);
  memcpy(msg->buf, argv[2], flen);
  
  if (SYSERR == writeToSocket(&nodes[peer].sock, 
			      &msg->header.header)) {
    fclose(infile);
    FREE(msg);
    PRINTF("ERROR: Could not send message to peer %s.\n", 
	   nodes[peer].ips);
    return -1;
  }
  /* Read ack from the peer */
  if (OK != readTCPResult(&nodes[peer].sock, &ack)) {
    fclose(infile);
    FREE(msg);
    PRINTF("Peer is not responding\n");
    return -1;
  }
  if (ack != OK) {
    fclose(infile);
    FREE(msg);
    PRINTF("ERROR: Peer returned error (delete existing file).\n");
    return -1;
  }
  msg->type = htonl(TESTBED_FILE_APPEND);
  buf = msg->buf + flen;
  while ((nbytes = GN_FREAD(buf, 1, 
			    (TESTBED_FILE_BLK_SIZE -
			     sizeof(TESTBED_UPLOAD_FILE_MESSAGE) - flen),
			    infile)) > 0) {
    if (ferror(infile))
      break;
    msg->header.header.size = htons(sizeof(TESTBED_UPLOAD_FILE_MESSAGE) +
				    nbytes + flen);
    if (SYSERR == writeToSocket(&nodes[peer].sock, &msg->header.header)) {
      fclose(infile);
      FREE(msg);
      PRINTF("ERROR: could not send file to node %s.\n",
	     nodes[peer].ips);
      return -1;
    }
    if (OK != readResult(peer, &ack)) {
      fclose(infile);
      FREE(msg);
      return -1;
    }
    if (ack != OK) {
      fclose(infile);
      FREE(msg);
      PRINTF("ERROR: peer returned error.\n");
      return -1;
    }
  }
  if (ferror(infile)) {
    fclose(infile);
    FREE(msg);
    PRINTF("ERROR: could not read source file. Transmission aborted.\n");
    return -1;
  }
  fclose(infile);
  FREE(msg);
  PRINTF("OK.\n");
  return 0;
}

/**
 * Print the list of commands.
 */
static int printOnlineHelp(int argc, 
			   char * argv[]) {
  int i;
  i = 0;
  while (commands[i].command != NULL) {
    PRINTF("%s\n",
	   commands[i].command);
    i++;
  }
  return 0;
}

/**
 * Print the list of available peers.
 */
static int listPeers(int argc, char * argv[]) {
  int i;
  for (i=0;i<nnodes;i++)
    PRINTF("%4d - %s:%d\n",
	   i,
	   nodes[i].ips,
	   nodes[i].port);
  return 0;
}

/**
 * Exit gnunet-testbed shell.
 */
static int doExit(int argc, char * argv[]) {
  do_quit = YES;
  return 0;
}

/* ****************** command set ****************** */

CMD_ENTRY commands[] = {
  { "help", &printOnlineHelp },
  { "get-trust", &getTrust },
  { "set-bw", &setBW },
  { "set-trust", &setTrust },
  { "add-node", &addNode },  
  { "connect", &addConnection },
  { "disconnect", &delConnection },
  { "disconnect-all", &delAllConnections },
  { "helo-disable", &disableHELO },
  { "helo-enable", &enableHELO },
  { "autoconnect-disable", &disableAUTOCONNECT },
  { "autoconnect-enable", &enableAUTOCONNECT },
  { "process-start", &startProcess },
  { "process-signal", &signalProcess },
  { "process-output", &dumpProcessOutput },
  { "exit", &doExit },
  { "list-peers", &listPeers}, 
  { "set-loss", &setLoss} ,
  { "get-stats", &getStats}, 
  { "get-stat", &getStat}, 
  { "load-module", &loadModule },
  { "unload-module", &unloadModule },
  { "upload", &uploadFile }, 
  { "connect-deny", &denyConnect },
  { "connect-allow", &allowConnect },
  { NULL, NULL }, /* termination */
};


/* end of commands.c */
